from importlib.util import module_from_spec, spec_from_file_location
import logging

import click

from .certs import generate_cert_files
from .constants import CADIR, DEFAULT_CN, DEFAULT_EXP, DEFAULT_ORG
from .proxy import Proxy, patch_mitmproxy_certfile_prefix


log = logging.getLogger(__name__)


def configure_logging(verbosity):
    level = logging.WARNING - verbosity * 10
    logging.basicConfig(format='[%(levelname)s] :: %(message)s', level=level)


def load_script(path):
    spec = spec_from_file_location('script', path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    log.info('Loaded script %s.', path)


@click.group()
@click.version_option(message='%(version)s')
@click.pass_context
def cli(ctx):
    """Padsniff: A Puzzle & Dragons HTTP Sniffer"""
    patch_mitmproxy_certfile_prefix()
    ctx.max_content_width = 120


@cli.command()
@click.option('--directory', '-d', default=CADIR, metavar='PATH',
              help='Directory to write files to. (default: ~/.padsniff)')
@click.option('--organization', '--org', '-o', default=DEFAULT_ORG,
              help='Organization name specified in the certificate signing request. (default: "padsniff")')
@click.option('--common-name', '--cn', '-c', default=DEFAULT_CN,
              help='Common name specified in the certificate signing request. (default: "Puzzle & Dragons HTTP Sniffer")')
@click.option('--expiry', '--exp', '-x', default=DEFAULT_EXP, metavar='SECONDS',
              help='Root certificate validity period in seconds. (default: 3 years)')
@click.option('--overwrite', '-f', is_flag=True,
              help='Overwrite certificate files if they exist.')
@click.option('--quiet', '-q', is_flag=True,
              help='Silence non-error output.')
def certs(directory, org, cn, exp, overwrite, quiet):
    """Generate a new root key and certificate."""
    filenames = generate_cert_files(directory, org, cn, exp, overwrite)

    if not quiet:
        for filename in filenames:
            click.echo(filename)


@cli.command()
@click.option('--cadir', '-d', default=CADIR, metavar='PATH',
              help='Path to directory containing certificate authority. (default: ~/.padsniff)')
@click.option('--port', '-p', default=8080,
              help='Proxy service port. (default: 8080)')
@click.option('--script', '-s', 'scripts', multiple=True, type=click.Path(exists=True),
              help='Load a Python script containing a callback. Can be supplied multiple times.')
@click.option('--verbose', '-v', 'verbosity', count=True,
              help='Increase logging level (default: WARNING). Can be supplied multiple times.')
def run(cadir, port, scripts, verbosity):
    """Run the proxy service."""
    configure_logging(verbosity)

    # load scripts before instantiating the proxy to allow the use of the `on` decorator
    for path in scripts:
        load_script(path)

    proxy = Proxy(host='0.0.0.0', port=port, cadir=cadir)
    proxy.run()



from queue import Queue
from threading import enumerate as all_threads, main_thread, Event, Thread
from time import sleep

from padsniff.parallel import parallelize, HandlerThread

from pytest import fixture


@fixture(autouse=True)
def join_threads():
    """
    Join all threads except the main thread to ensure none
    are running between tests.
    """
    threads = set(all_threads()) - {main_thread()}
    for thread in threads:
        thread.join()


class TestHandlerThread:


    def test_instance_tracking(self):
        """Test that HandlerThread keeps track of instances of itself."""
        assert len(HandlerThread.instances) == 0

        handler_thread1 = HandlerThread()
        handler_thread2 = HandlerThread()
        regular_thread = Thread()

        assert len(HandlerThread.instances) == 2
        assert handler_thread1 in HandlerThread.instances
        assert handler_thread2 in HandlerThread.instances
        assert regular_thread not in HandlerThread.instances

        # HandlerThread.instances is a WeakSet, so allowing the gc to
        # collect handler_thread2 will remove it from the set
        del handler_thread2

        assert len(HandlerThread.instances) == 1


def test_parallelize_nonblocking():
    """Test basic functionality of parallelize decorator."""
    # only main thread should be running
    assert len(all_threads()) == 1
    event = Event()

    def wait_until_event():
        event.wait()

    # test that decorator sets the __wrapped__ attribute
    # mimicking functools.wraps' behavior
    assert parallelize(wait_until_event).__wrapped__ is wait_until_event

    # test that decorated function is run without blocking the main thread
    parallelize(wait_until_event)()

    assert len(all_threads()) == 2

    event.set()
    # give the thread time to wake up and exit
    sleep(0.1)

    assert len(all_threads()) == 1


def test_parallellize_with_args(mocker):
    """Test that parallelize decorator passes arguments to the target function."""
    mock = mocker.Mock()
    name = mock.__qualname__ = 'mock'
    args = 'can dank memes', 'melt steel beams?'
    kwargs = dict(tests='pls pass')
    parallel_func = parallelize(mock)

    parallel_func(*args, **kwargs)

    assert mock.called_once_with(
        target=mock,
        name=name,
        args=args,
        kwargs=kwargs,
    )
    assert mock.start.called_once()


def test_parallelize_thread_safe_args():
    """Tests that arguments passed to parallelize-decorated function are thread-safe."""
    queue = Queue()
    mutable_arg = []
    mutable_kwarg = {}

    @parallelize
    def func(arg, kwarg):
        queue.put(arg)
        queue.put(kwarg)

    func(mutable_arg, kwarg=mutable_kwarg)
    thread_arg = queue.get()
    thread_kwarg = queue.get()

    assert thread_arg == mutable_arg
    assert thread_arg is not mutable_arg
    assert thread_kwarg == mutable_kwarg
    assert thread_kwarg is not mutable_kwarg



from mitmproxy.test.tflow import tflow as TestFlow
from pytest import fixture
from pytest_mock import mocker


class MockSocket:
    """A Mock `socket.socket` to prevent opening connections."""

    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, method_name):
        # mock all method names to return called parameters
        # prevents sockets from binding and connecting
        return lambda *args, **kwargs: (args, kwargs)


@fixture(autouse=True)
def mock_proxy_server(monkeypatch):
    """Prevent networking libraries that use `socket.socket` from opening connections."""
    monkeypatch.setattr('socket.socket', MockSocket)


@fixture
def flow():
    return TestFlow()



from unittest.mock import patch

from pytest import fixture

from padsniff import constants, is_gungho, on, parallelize, Proxy


@fixture(autouse=True)
def clear_Proxy_handlers():
    """Clear `Proxy.handlers` before each test case to prevent handlers from persisting."""
    Proxy.handlers.clear()


class TestProxy:


    def test_init(self):
        """Test initialization of Proxy instance with custom options."""
        host, port, cadir = '1.2.3.4', 12345, '~/.padsniff'

        with patch('padsniff.proxy.generate_cert_files') as mocked:
            proxy = Proxy(host=host, port=port, cadir=cadir)

        assert mocked.called_once_with(cadir)

        # test that parent class initializes with proper arguments
        assert proxy.options.listen_host == host
        assert proxy.options.listen_port == port
        assert proxy.options.mode == 'transparent'

        # test that the handlers are equivalent, but not the same reference
        assert proxy.handlers == Proxy.handlers
        assert proxy.handlers is not Proxy.handlers


    def test_inst_on_decorator(self, mocker):
        """Test basic functionality of the class-level on decorator."""
        endpoint = 'endpoint'
        proxy = Proxy()

        # test that instance decorator adds only to instance variable
        @proxy.on(endpoint, blocking=True)
        def func():
            pass

        assert func in proxy.handlers[endpoint]
        assert func not in Proxy.handlers[endpoint]

        # test that padsniff.parallel.parallelize is only called for blocking=False
        with mocker.patch('padsniff.parallel.parallelize') as mock:
            proxy.on(endpoint, blocking=True)(func)
            assert mock.not_called()

            proxy.on(endpoint, blocking=False)(func)
            assert mock.called_once_with(func)

        # test that decorated function is returned invariant
        assert proxy.on(endpoint, blocking=True)(func) is func
        assert proxy.on(endpoint, blocking=False)(func) is func


    def test_response_filtering(self, mocker, flow):
        """Test capturing only flows that match GungHo's user agent and endpoint URL."""
        route = mocker.patch('padsniff.proxy.Proxy.route')
        proxy = Proxy()

        flow.request.headers['user-agent'] = constants.GUNGHO_USER_AGENT
        flow.request.path = constants.GUNGHO_API_ENDPOINT
        proxy.response(flow)
        assert route.called_once_with(flow)

        route.reset_mock()
        flow.request.headers['user-agent'] = 'PadsniffTesting'
        flow.request.path = constants.GUNGHO_API_ENDPOINT
        proxy.response(flow)
        assert route.not_called()

        route.reset_mock()
        flow.request.headers['user-agent'] = constants.GUNGHO_USER_AGENT
        flow.request.path = '/path/to/success'
        proxy.response(flow)
        assert route.not_called()

        route.reset_mock()
        flow.request.headers['user-agent'] = 'PadsniffTesting'
        flow.request.path = '/path/to/success'
        proxy.response(flow)
        assert route.not_called()


    def test_flow_routing(self, mocker, flow):
        """Test routing flows to the correct handlers."""
        endpoint = 'endpoint1'
        func1 = mocker.Mock()
        func2 = mocker.Mock()
        proxy = Proxy()

        proxy.on(endpoint)(func1)
        proxy.on('endpoint2')(func2)

        flow.request.path = '%s?action=%s' % (constants.GUNGHO_API_ENDPOINT, endpoint)
        proxy.route(flow)

        assert func1.called_once_with(flow.request, flow.response)
        assert func2.not_called()


def test_on_decorator(mocker):
    """Test basic functionality of the module-level on decorator."""
    endpoint = 'endpoint'
    proxy = Proxy()

    # test that module-level decorator adds only to class variable
    @on(endpoint, blocking=True)
    def func():
        pass

    assert func in Proxy.handlers[endpoint]
    assert func not in proxy.handlers[endpoint]

    # test that padsniff.parallel.parallelize is only called for blocking=False
    with mocker.patch('padsniff.parallel.parallelize') as mock:
        on(endpoint, blocking=True)(func)
        assert mock.not_called(func)

        on(endpoint, blocking=False)(func)
        assert mock.called_once_with(func)

    # test that decorated function is returned invariant
    assert on(endpoint, blocking=True)(func) is func
    assert on(endpoint, blocking=False)(func) is func

    # test that an arbitrary class can be specified
    class NewProxy:
        on = mocker.Mock()

    on(endpoint, blocking=True, cls=NewProxy)

    assert NewProxy.on.called_once_with(endpoint, blocking=True)


def test_is_gungho(flow):
    flow.request.headers['user-agent'] = constants.GUNGHO_USER_AGENT
    flow.request.path = constants.GUNGHO_API_ENDPOINT
    assert is_gungho(flow.request)

    flow.request.headers['user-agent'] = 'PadsniffTesting'
    flow.request.path = constants.GUNGHO_API_ENDPOINT
    assert not is_gungho(flow.request)

    flow.request.headers['user-agent'] = constants.GUNGHO_USER_AGENT
    flow.request.path = '/path/to/success'
    assert not is_gungho(flow.request)

    flow.request.headers['user-agent'] = 'PadsniffTesting'
    flow.request.path = '/path/to/success'
    assert not is_gungho(flow.request)



from pathlib import Path

from mitmproxy.certs import create_ca
from OpenSSL.crypto import dump_certificate, dump_privatekey, FILETYPE_PEM

from .constants import DEFAULT_CN, DEFAULT_EXP, DEFAULT_ORG


def generate_root_pair(organization, common_name, expiry):
    """
    Creates a new root key and certificate.

    The pair is returned as a two-tuple of bytes objects (key, ca) in PEM format.
    """
    key, ca = create_ca(o=organization, cn=common_name, exp=expiry)
    key_bytes = dump_privatekey(FILETYPE_PEM, key)
    ca_bytes = dump_certificate(FILETYPE_PEM, ca)
    return key_bytes, ca_bytes


def generate_cert_files(directory, organization=DEFAULT_ORG, common_name=DEFAULT_CN, expiry=DEFAULT_EXP, overwrite=False):
    """
    Creates a new root key and certificate, and dumps them to file in PEM format.

    The following files are created:
        padsniff-ca-key.pem, containing the root key.
        padsniff-ca-cert.pem, containing the root certificate.
        padsniff-ca.pem, containing the root key and certificate together.

    The written filenames are returned as `pathlib.Path` objects, in the order
    enumerated above.
    """
    path = Path(directory).expanduser().resolve()
    path.mkdir(parents=True, exist_ok=True)

    keyfile = path / (DEFAULT_ORG + '-ca-key.pem')
    certfile = path / (DEFAULT_ORG + '-ca-cert.pem')
    pairfile = path / (DEFAULT_ORG + '-ca.pem')
    filenames = keyfile, certfile, pairfile

    if not overwrite and any(f.exists() for f in filenames):
        return ()

    key, ca = generate_root_pair(organization, common_name, expiry)

    # mitmproxy also generates Diffie-Hellman parameters, but this is
    # both unsupported by pyOpenSSL (as of v16.2.0), and takes forever
    # to generate with `openssl dhparam`, so we'll just let mitmproxy
    # handle it

    with open(keyfile, 'wb') as f:
        f.write(key)

    with open(certfile, 'wb') as f:
        f.write(ca)

    with open(pairfile, 'wb') as f:
        f.write(key)
        f.write(ca)

    return filenames



from padsniff.structures import CaseInsensitiveDefaultDict

import pytest


class TestCaseInsensitiveDefaultDict:


    def test_get_set_delitem(self):
        d = CaseInsensitiveDefaultDict(default_factory=None)
        key = 'key'
        value = object()

        d[key] = value

        assert d[key] is value

        del d[key]

        assert key not in d


    def test_contains(self):
        d = CaseInsensitiveDefaultDict(default_factory=None)
        key = 'key'

        assert key not in d

        d[key] = object()

        assert key in d


    def test_case_insensitivity(self):
        d = CaseInsensitiveDefaultDict(default_factory=None)
        key = 'KeY'
        d[key] = object()

        assert d[key] is d[key.upper()]
        assert d[key] is d[key.lower()]
        assert d[key] is d[key.casefold()]


    def test_original_capitalization(self):
        d = CaseInsensitiveDefaultDict(default_factory=None)
        key = 'KeY'
        d[key] = object()
        d_as_dict = dict(d)

        assert key in d_as_dict
        assert key.upper() not in d_as_dict
        assert key.lower() not in d_as_dict
        assert key.casefold() not in d_as_dict


    def test_key_transform(self):
        transform = CaseInsensitiveDefaultDict._transform

        assert transform('StRiNg') == 'string'
        assert transform(1) == 1
        assert transform(range(10)) == range(10)
        assert transform(set('AbC')) == set('AbC')

        obj = object()

        assert transform(obj) is obj


    def test_valid_default_factory(self, mocker):
        mock_default_factory = mocker.Mock()
        mock_default_factory.return_value = object()

        d = CaseInsensitiveDefaultDict(default_factory=mock_default_factory)
        key = 'key'

        assert d.default_factory is mock_default_factory
        assert d[key] == mock_default_factory.return_value
        assert mock_default_factory.called_once_with(key)


    def test_non_callable_default_factory(self):
        with pytest.raises(TypeError):
            CaseInsensitiveDefaultDict(default_factory=True)


    def test_no_default_factory(self):
        d = CaseInsensitiveDefaultDict(default_factory=None)
        key = 'key'

        assert d.default_factory is None
        assert key not in d

        with pytest.raises(KeyError):
            d[key]


    def test_repr(self):
        d = CaseInsensitiveDefaultDict(default_factory=int)
        assert repr(d) == 'CaseInsensitiveDefaultDict(int, {})'

        d.update(A=1)
        assert repr(d) == "CaseInsensitiveDefaultDict(int, {'A': 1})"


    def test_copy(self):
        d = CaseInsensitiveDefaultDict(default_factory=None, a=1, b=2, c=3)

        assert d == d.copy()


    def test_get(self, mocker):
        mock_default_factory = mocker.Mock()
        d = CaseInsensitiveDefaultDict(default_factory=mock_default_factory)
        key = 'key'
        obj = object()
        d[key] = obj

        assert d.get(key) is obj
        assert mock_default_factory.not_called()

        key2 = 'key2'
        obj2 = object()

        assert d.get(key2) is None
        assert d.get(key2, obj2) is obj2
        assert key2 not in d
        assert mock_default_factory.not_called()


    def test_setdefault(self, mocker):
        mock_default_factory = mocker.Mock()
        d = CaseInsensitiveDefaultDict(default_factory=mock_default_factory)
        key = 'key'
        obj = object()
        d[key] = obj

        assert d.setdefault(key) is obj
        assert mock_default_factory.not_called()

        key2 = 'key2'
        obj2 = object()

        assert d.setdefault(key2) is None
        assert d.setdefault(key2, obj2) is None
        assert key2 in d
        assert mock_default_factory.not_called()



    def test_fromkeys(self):
        obj = object()
        d = CaseInsensitiveDefaultDict.fromkeys(default_factory=None, seq='abc', value=obj)

        assert d.default_factory is None
        assert d == CaseInsensitiveDefaultDict(None, a=obj, b=obj, c=obj)



from mitmproxy.certs import DEFAULT_EXP

ALL = '*'

CADIR = '~/.padsniff'
DEFAULT_CN = 'Puzzle & Dragons HTTP Sniffer'
DEFAULT_ORG = 'padsniff'

GUNGHO_API_ENDPOINT = '/api.php'
GUNGHO_USER_AGENT = 'GunghoPuzzleAndDungeon'



from contextlib import suppress
from os import remove
from os.path import dirname, join

from OpenSSL.crypto import load_certificate, FILETYPE_PEM
import pytest

from padsniff.certs import generate_cert_files


FILENAMES = 'padsniff-ca.pem', 'padsniff-ca-cert.pem', 'padsniff-ca-key.pem'
HERE = dirname(__file__)


@pytest.fixture(autouse=True)
def cleanup():
    """Delete generated files after each test."""
    yield

    for filename in FILENAMES:
        path = join(HERE, filename)

        with suppress(FileNotFoundError):
            remove(path)


class TestCertGeneration:


    def test_subject(self):
        """Test certificate subject customization."""
        org, cn = 'test org', 'test name'
        _, certfile, _ = generate_cert_files(HERE, organization=org, common_name=cn)

        with open(certfile, 'rb') as f:
            certdata = f.read()

        cert = load_certificate(FILETYPE_PEM, certdata)
        subject = cert.get_subject()
        assert subject.O == org
        assert subject.CN == cn


    def test_no_overwrite(self):
        """Test certificate generation does not overwrite files by default."""
        expected = {join(HERE, filename) for filename in FILENAMES}

        generated = {str(path) for path in generate_cert_files(HERE)}
        assert generated == expected

        generated2 = generate_cert_files(HERE)
        assert not generated2


    def test_overwrite(self):
        """Test certificate generation does overwrite when specified."""
        expected = {join(HERE, filename) for filename in FILENAMES}

        generated = {str(path) for path in generate_cert_files(HERE)}
        assert generated == expected

        generated2 = {str(path) for path in generate_cert_files(HERE, overwrite=True)}
        assert set(generated2) == expected
